from __future__ import annotations

from collections.abc import Callable
from re import Pattern
from typing import Generic, TypeVar

from arclet.cithun.config import Config
from arclet.cithun.exceptions import PermissionDeniedError, ResourceNotFoundError
from arclet.cithun.model import Permission, ResourceNode, Role, User

from .service import AsyncPermissionService
from .store import AsyncStore

T = TypeVar("T")


class AsyncPermissionExecutor(Generic[T]):
    """
    专门负责对“权限状态”的 get/set 操作。

    这里的“权限状态”定义为：某个 subject 在某个资源节点上的有效权限 bitmask（int）。

    - suget / suset : 以 root 身份操作（无权限检查）
    - get / set     : 以执行者（某个 User）身份操作，并按题设规则做检查
    """

    def __init__(
        self,
        storage: AsyncStore,
        perm_service: AsyncPermissionService[T],
    ):
        self.storage = storage
        self.service = perm_service

    async def _get_parent_and_self(
        self,
        resource_path: str,
        missing_ok: bool,
    ) -> tuple[ResourceNode | None, ResourceNode | None]:
        """获取父节点和当前节点。

        Args:
            resource_path (str): 资源路径。
            missing_ok (bool): 是否允许节点不存在。

        Returns:
            tuple[ResourceNode | None, ResourceNode | None]: (parent_node, self_node)。

        Raises:
            ResourceNotFoundError: 当 missing_ok=False 且节点不存在时抛出。
        """
        path = resource_path.strip(Config.NODE_SEPARATOR)
        if not path:
            # 根节点没有父节点的概念，这里约定 parent=None, self=root(如果存在)
            # root = self.storage.resources.get("root")
            # if root is None and not missing_ok:
            if missing_ok:
                return None, None
            raise ResourceNotFoundError("Root resource 'root' does not exist")

        # 获取 self
        try:
            self_node = await self.storage.get_resource(resource_path)
        except KeyError:
            self_node = None

        # 获取 parent
        parent_id = None
        if Config.NODE_SEPARATOR in path:
            parent_id, _, _ = path.rpartition(Config.NODE_SEPARATOR)

        try:
            parent_node = await self.storage.get_resource(parent_id) if parent_id else None
        except KeyError:
            parent_node = None

        # 父节点不存在
        if parent_node is None and parent_id and not missing_ok:
            raise ResourceNotFoundError(f"Parent resource '{parent_id}' not found")

        # 自身不存在
        if self_node is None and not missing_ok:
            raise ResourceNotFoundError(f"Resource '{path}' not found")

        return parent_node, self_node

    async def _ensure_resource_for_set(
        self,
        resource_path: str,
        missing_ok: bool,
    ) -> tuple[ResourceNode | None, ResourceNode]:
        """确保资源存在以进行设置操作。

        Args:
            resource_path (str): 资源路径。
            missing_ok (bool): 是否允许自动创建。

        Returns:
            tuple[ResourceNode | None, ResourceNode]: (parent_node, self_node)。

        Raises:
            ResourceNotFoundError: 当 missing_ok=False 且资源不存在时抛出。
        """
        parent, node = await self._get_parent_and_self(resource_path, missing_ok=missing_ok)
        if node is None:
            if not missing_ok:
                raise ResourceNotFoundError(f"Resource '{resource_path}' not found")
            # 自动创建
            node = await self.storage.define(resource_path)
        return parent, node

    def _apply_chmod(
        self,
        old_mask: Permission,
        mask: Permission,
        mode: str,
    ) -> Permission:
        """应用 chmod 风格的权限调整。

        Args:
            old_mask (int): 旧的权限掩码。
            mask (int): 新的权限掩码。
            mode (str): 调整模式。
                "=" : 覆盖为 mask
                "+" : old_mask | mask
                "-" : old_mask & ~mask

        Returns:
            int: 调整后的权限掩码。

        Raises:
            ValueError: 当 mode 不支持时抛出。
        """
        if mode == "=":
            return mask
        elif mode == "+":
            return old_mask | mask
        elif mode == "-":
            return old_mask & ~mask
        else:
            raise ValueError(f"Unsupported chmod mode: {mode!r}")

    async def suget(
        self,
        subject: User | Role,
        resource_path: str,
        missing_ok: bool = False,
        context: T | None = None,
    ) -> Permission | None:
        """root 获取 subject 在指定节点上的权限状态（bitmask），不做权限校验。

        Args:
            subject (User | Role): 目标主体（用户或角色）。
            resource_path (str): 资源路径。
            missing_ok (bool, optional): 是否允许节点不存在。默认为 False。
            context (T | None, optional): 上下文信息。默认为 None。

        Returns:
            Permission | None: 权限掩码。如果 missing_ok=True 且节点不存在，返回 None。

        Raises:
            ResourceNotFoundError: 当 missing_ok=False 且节点不存在时抛出。
        """
        _, node = await self._get_parent_and_self(resource_path, missing_ok=missing_ok)
        # 若 missing_ok=True，则 parent 或 node 可能为 None，此时按规则返回 None
        if node is None:
            return None

        # 直接用内部方法计算该 subject 的静态权限
        if isinstance(subject, Role):
            cache: dict[tuple[str, str, str], Permission] = {}
            mask = await self.service._calc_permissions_for_subject(
                subject.type,
                subject.id,
                node,
                context,
                visited=[],
                cache=cache,
            )
        else:
            mask = await self.service.get_effective_permissions(subject, node.id, context)
        return mask

    async def test(
        self,
        subject: User | Role,
        resource_path: str,
        required_mask: Permission,
        missing_ok: bool = False,
        context: T | None = None,
    ) -> bool:
        """root 测试 subject 在指定节点上是否拥有某些权限，不做权限校验。

        Args:
            subject (User | Role): 目标主体（用户或角色）。
            resource_path (str): 资源路径。
            required_mask (Permission): 需要的权限掩码。
            missing_ok (bool, optional): 是否允许节点不存在。默认为 False。
            context (T | None, optional): 上下文信息。默认为 None。

        Returns:
            bool: 是否拥有指定权限。如果 missing_ok=True 且节点不存在，返回 False。

        Raises:
            ResourceNotFoundError: 当 missing_ok=False 且节点不存在时抛出。
        """
        mask = await self.suget(
            subject,
            resource_path,
            missing_ok=missing_ok,
            context=context,
        )
        if mask is None:
            mask = Permission.VISIT | Permission.AVAILABLE
        return (mask & required_mask) == required_mask

    async def get(
        self,
        executor: User,
        resource_path: str,
        missing_ok: bool = False,
        context: T | None = None,
    ) -> Permission | None:
        """执行者获取自己在目标节点的权限状态。

        Args:
            executor (User): 执行者。
            resource_path (str): 资源路径。
            missing_ok (bool, optional): 是否允许节点不存在。默认为 False。
            context (tuple | None, optional): 上下文信息。默认为 None。

        Returns:
            Permission | None: 权限掩码。如果 missing_ok=True 且节点不存在，返回 None。

        Raises:
            ResourceNotFoundError: 当 missing_ok=False 且节点不存在时抛出。
            PermissionDeniedError: 当执行者在节点缺少 VISIT 权限时抛出
        """
        _, node = await self._get_parent_and_self(resource_path, missing_ok=missing_ok)

        # 3. 自身不存在
        if node is None:
            return None

        # 4. 检查自身权限包含 VISIT
        # 如果某个父节点显式 deny V，这里自然拿不到 V
        self_mask = await self.service.get_effective_permissions(executor, node.id, context)
        if (self_mask & Permission.VISIT) != Permission.VISIT:
            raise PermissionDeniedError(f"Executor '{executor.id}' lacks VISIT on '{node.id}'")
        return self_mask

    async def suset(
        self,
        subject: User | Role,
        resource_path: str | Callable[[str], bool] | Pattern[str],
        mask: Permission,
        mode: str = "=",
        deny: bool = False,
        missing_ok: bool = False,
    ) -> None:
        """root 为 subject 在指定节点上“设置权限状态”。

        Args:
            subject (User | Role): 目标主体（用户或角色）。
            resource_path (str | Callable[[str], bool] | Pattern[str]): 资源路径或匹配模式。
            mask (Permission): 权限掩码。
            mode (str, optional): 设置模式 ("=", "+", "-")。默认为 "="。
            deny (bool, optional): 是否设置为 deny 掩码。默认为 False（设置为 allow）。
            missing_ok (bool, optional): 是否允许节点不存在。默认为 False。

        Raises:
            ResourceNotFoundError: 当 missing_ok=False 且节点不存在时抛出。
            ValueError: 当 mode 不支持时抛出。
        """

        if isinstance(resource_path, str):
            resource_path = resource_path.strip(Config.NODE_SEPARATOR)
            if "*" in resource_path or "?" in resource_path or "[" in resource_path:
                matched = await self.storage.glob_resources(resource_path)
            else:
                _, node = await self._ensure_resource_for_set(resource_path, missing_ok=missing_ok)

                primary_acl = await self.storage.get_primary_acl(subject, node.id)
                old_mask = (
                    (primary_acl.deny_mask if deny else primary_acl.allow_mask) if primary_acl else Permission.NONE
                )
                new_mask = self._apply_chmod(old_mask, mask, mode)
                if primary_acl is None:
                    await self.storage.assign(
                        subject=subject,
                        resource_path=node.id,
                        allow_mask=Permission.NONE if deny else new_mask,
                        deny_mask=new_mask if deny else Permission.NONE,
                    )
                else:
                    await self.storage.update_acl(
                        primary_acl,
                        allow_mask=new_mask if not deny else primary_acl.allow_mask,
                        deny_mask=new_mask if deny else primary_acl.deny_mask,
                    )
                return
        elif isinstance(resource_path, Pattern):
            matched = await self.storage.match_resources(lambda t: bool(resource_path.fullmatch(t)))
        else:
            matched = await self.storage.match_resources(resource_path)
        for node in matched:
            if node.parent_id and node.parent_id not in self.storage.resources:
                if not missing_ok:
                    raise ResourceNotFoundError(f"Parent '{node.parent_id}' for '{node.id}' not found")
                else:
                    continue

            primary_acl = await self.storage.get_primary_acl(subject, node.id)
            old_mask = (primary_acl.deny_mask if deny else primary_acl.allow_mask) if primary_acl else Permission.NONE
            new_mask = self._apply_chmod(old_mask, mask, mode)
            if primary_acl is None:
                await self.storage.assign(
                    subject=subject,
                    resource_path=node.id,
                    allow_mask=Permission.NONE if deny else new_mask,
                    deny_mask=new_mask if deny else Permission.NONE,
                )
            else:
                await self.storage.update_acl(
                    primary_acl,
                    allow_mask=new_mask if not deny else primary_acl.allow_mask,
                    deny_mask=new_mask if deny else primary_acl.deny_mask,
                )

    async def chmod(
        self,
        subject: User | Role,
        resource_path: str | Callable[[str], bool] | Pattern[str],
        pattern: str,
        missing_ok: bool = False,
    ) -> None:
        """root 为 subject 在指定节点上“设置权限状态”。

        Args:
            subject (User | Role): 目标主体（用户或角色）。
            resource_path (str | Callable[[str], bool] | Pattern[str]): 资源路径或匹配模式。
            pattern (str): 权限表达式。形如 [target][op][flags]。
            missing_ok (bool, optional): 是否允许节点不存在。默认为 False。

        Raises:
            ResourceNotFoundError: 当 missing_ok=False 且节点不存在时抛出。
            ValueError: 当 mode 不支持时抛出。
        """
        return await self.suset(subject, resource_path, *Permission.parse(pattern), missing_ok=missing_ok)

    async def set(
        self,
        executor: User,
        target: User | Role,
        resource_path: str | Callable[[str], bool] | Pattern[str],
        mask: Permission,
        mode: str = "=",
        deny: bool = False,
        missing_ok: bool = False,
        context: T | None = None,
    ) -> None:
        """执行者为目标 subject 设置目标节点的权限状态。

        Args:
            executor (User): 执行者。
            target (User | Role): 目标主体（用户或角色）。
            resource_path (str | Callable[[str], bool] | Pattern[str]): 资源路径或匹配模式。
            mask (Permission): 权限掩码。
            mode (str, optional): 设置模式 ("=", "+", "-")。默认为 "="。
            deny (bool, optional): 是否设置为 deny 掩码。默认为 False（设置为 allow）。
            missing_ok (bool, optional): 是否允许节点不存在。默认为 False。
            context (tuple | None, optional): 上下文信息。默认为 None。

        Raises:
            ResourceNotFoundError: 当 missing_ok=False 且节点不存在时抛出。
            PermissionDeniedError: 当执行者权限不足时抛出。
            ValueError: 当 mode 不支持时抛出。
        """
        if isinstance(resource_path, str):
            resource_path = resource_path.strip(Config.NODE_SEPARATOR)
            if "*" in resource_path or "?" in resource_path or "[" in resource_path:
                matched = await self.storage.glob_resources(resource_path)
            else:
                parent, node = await self._ensure_resource_for_set(resource_path, missing_ok=missing_ok)

                # 2. 检查执行者在父节点是否有 v+m+a
                if parent is not None:
                    parent_mask = await self.service.get_effective_permissions(executor, parent.id, context)
                    required_parent = Permission.VISIT | Permission.MODIFY | Permission.AVAILABLE
                    if (parent_mask & required_parent) != required_parent:
                        raise PermissionDeniedError(f"Executor '{executor.id}' lacks V+M+A on parent '{parent.id}'")

                # 4. 检查执行者在自身是否有 MODIFY
                self_mask = await self.service.get_effective_permissions(executor, node.id, context)
                if (self_mask & Permission.MODIFY) != Permission.MODIFY:
                    # 不修改该节点状态
                    return

                primary_acl = await self.storage.get_primary_acl(target, node.id)
                old_mask = (
                    (primary_acl.deny_mask if deny else primary_acl.allow_mask) if primary_acl else Permission.NONE
                )
                new_mask = self._apply_chmod(old_mask, mask, mode)
                if primary_acl is None:
                    await self.storage.assign(
                        subject=target,
                        resource_path=node.id,
                        allow_mask=Permission.NONE if deny else new_mask,
                        deny_mask=new_mask if deny else Permission.NONE,
                    )
                else:
                    await self.storage.update_acl(
                        primary_acl,
                        allow_mask=new_mask if not deny else primary_acl.allow_mask,
                        deny_mask=new_mask if deny else primary_acl.deny_mask,
                    )
                return
        elif isinstance(resource_path, Pattern):
            matched = await self.storage.match_resources(lambda t: bool(resource_path.fullmatch(t)))
        else:
            matched = await self.storage.match_resources(resource_path)
        for node in matched:
            parent = self.storage.resources.get(node.parent_id) if node.parent_id else None
            if node.parent_id and parent is None:
                if not missing_ok:
                    raise ResourceNotFoundError(f"Parent '{node.parent_id}' for '{node.id}' not found")
                else:
                    continue
            if parent is not None:
                parent_mask = await self.service.get_effective_permissions(executor, parent.id, context)
                required_parent = Permission.VISIT | Permission.MODIFY | Permission.AVAILABLE
                if (parent_mask & required_parent) != required_parent:
                    raise PermissionDeniedError(f"Executor '{executor.id}' lacks V+M+A on parent '{parent.id}'")
            self_mask = await self.service.get_effective_permissions(executor, node.id, context)
            if (self_mask & Permission.MODIFY) != Permission.MODIFY:
                continue  # 无修改权限，跳过

            # 5. chmod 更新目标 subject
            primary_acl = await self.storage.get_primary_acl(target, node.id)
            old_mask = (primary_acl.deny_mask if deny else primary_acl.allow_mask) if primary_acl else Permission.NONE
            new_mask = self._apply_chmod(old_mask, mask, mode)
            if primary_acl is None:
                await self.storage.assign(
                    subject=target,
                    resource_path=node.id,
                    allow_mask=Permission.NONE if deny else new_mask,
                    deny_mask=new_mask if deny else Permission.NONE,
                )
            else:
                await self.storage.update_acl(
                    primary_acl,
                    allow_mask=new_mask if not deny else primary_acl.allow_mask,
                    deny_mask=new_mask if deny else primary_acl.deny_mask,
                )



from .exceptions import DependencyCycleError as DependencyCycleError
from .exceptions import PermissionDeniedError as PermissionDeniedError
from .exceptions import ResourceNotFoundError as ResourceNotFoundError
from .function import PermissionExecutor as PermissionExecutor
from .model import InheritMode as InheritMode
from .model import Permission as Permission
from .model import ResourceNode as ResourceNode
from .model import Role as Role
from .model import User as User
from .service import PermissionService as PermissionService
from .store import BaseStore as BaseStore
from .strategy import PermissionEngine as PermissionEngine

PE = PermissionExecutor
PermissionNotFoundError = ResourceNotFoundError



from __future__ import annotations

import json
import os
from dataclasses import asdict
from pathlib import Path

from arclet.cithun.model import AclDependency, AclEntry, ResourceNode, Role, Track, TrackLevel, User
from arclet.cithun.store import BaseStore


class JsonStore(BaseStore):
    def load(self):
        if self.file.exists():
            with self.file.open("r", encoding="utf-8") as f:
                data = json.load(f)
            user_rows = data.get("users", [])
            for row in user_rows:
                user = User(**row)
                self.users[user.id] = user
            role_rows = data.get("roles", [])
            for row in role_rows:
                role = Role(**row)
                self.roles[role.id] = role
            resource_rows = data.get("resources", [])
            for row in resource_rows:
                resource = ResourceNode(**row)
                self.resources[resource.id] = resource

            acl_rows = data.get("acls", [])
            for row in acl_rows:
                dependencies = row.pop("dependencies", [])
                acl = AclEntry(**row)
                for dep_row in dependencies:
                    dep = AclDependency(**dep_row)
                    acl.dependencies.append(dep)
                self.acls.append(acl)
            track_rows = data.get("tracks", [])
            for row in track_rows:
                levels = row.pop("levels", [])
                track = Track(**row)
                for level_row in levels:
                    level = TrackLevel(**level_row)
                    track.levels.append(level)
                self.tracks[track.id] = track
        else:
            with self.file.open("w+", encoding="utf-8") as f:
                json.dump({}, f, ensure_ascii=False)

    def save(self):
        data = {
            "users": [asdict(user) for user in self.users.values()],
            "roles": [asdict(role) for role in self.roles.values()],
            "resources": [asdict(res) for res in self.resources.values()],
            "acls": [asdict(acl) for acl in self.acls],
            "tracks": [asdict(track) for track in self.tracks.values()],
        }

        with self.file.open("w", encoding="utf-8") as f:
            json.dump(data, f, ensure_ascii=False, indent=2)

    def __init__(self, file: os.PathLike[str]):
        file = Path(file)
        if not file.suffix.startswith(".json"):
            raise ValueError(file)
        self.file = file
        super().__init__()



from __future__ import annotations

from typing import Generic, TypeVar

from .exceptions import DependencyCycleError
from .model import AclEntry, InheritMode, Permission, ResourceNode, Role, SubjectType, User
from .store import BaseStore
from .strategy import PermissionEngine

T = TypeVar("T")


def expand_roles(role_ids: list[str], roles: dict[str, Role]) -> set[str]:
    """递归展开角色列表，获取所有继承的角色 ID。

    Args:
        role_ids (list[str]): 初始角色 ID 列表。
        roles (dict[str, Role]): 角色字典，用于查找角色信息。

    Returns:
        set[str]: 包含所有继承角色的 ID 集合。
    """
    result: set[str] = set()
    visited: set[str] = set()

    def dfs(rid: str):
        if rid in visited:
            return
        visited.add(rid)

        role = roles.get(rid)
        if not role:
            return

        result.add(rid)
        for pr in role.parent_role_ids:
            dfs(pr)

    for rid in role_ids:
        dfs(rid)
    return result


class PermissionService(Generic[T]):
    def __init__(self, storage: BaseStore, engine: PermissionEngine[T]):
        self.storage = storage
        self.engine = engine

    def get_effective_permissions(
        self,
        user: str | User,
        resource_id: str,
        context: T | None = None,
    ) -> Permission:
        """计算用户在指定资源上的有效权限。

        Args:
            user (str | User): 用户 ID 或用户对象。
            resource_id (str): 资源 ID。
            context (T | None, optional): 上下文信息。默认为 None。

        Returns:
            Permission: 有效权限掩码。
        """
        resource = self.storage.get_resource(resource_id)
        user_id = user.id if isinstance(user, User) else user
        cache: dict[tuple[str, str, str], Permission] = {}

        def permission_lookup(subject: User | Role, ctx: T | None) -> Permission:
            return self._calc_permissions_for_subject(subject.type, subject.id, resource, ctx, visited=[], cache=cache)

        base_mask = self._calc_permissions_for_subject(
            SubjectType.USER, user_id, resource, context, visited=[], cache=cache
        )

        return self.engine.apply_strategies(
            self.storage.get_user(user_id),
            resource,
            context,
            base_mask,
            permission_lookup,
        )

    def has_permission(
        self,
        user: str | User,
        resource_id: str,
        required_mask: int,
        context: T | None = None,
    ) -> bool:
        """检查用户是否拥有指定资源的特定权限。

        Args:
            user (str | User): 用户 ID 或用户对象。
            resource_id (str): 资源 ID。
            required_mask (int): 需要的权限掩码。
            context (T | None, optional): 上下文信息。默认为 None。

        Returns:
            bool: 如果拥有所有请求的权限则返回 True，否则返回 False。
        """
        eff = self.get_effective_permissions(user, resource_id, context)
        return (eff & required_mask) == required_mask

    def _calc_permissions_for_subject(
        self,
        subject_type: SubjectType,
        subject: str | User | Role,
        resource: ResourceNode,
        context: T | None,
        visited: list[tuple[str, str, str]],
        cache: dict[tuple[str, str, str], Permission],
    ) -> Permission:
        subject_id = subject.id if isinstance(subject, (User, Role)) else subject
        key = (subject_type.value, subject_id, resource.id)
        if key in cache:
            return cache[key]
        if key in visited:
            cycle_start = visited.index(key)
            cycle = visited[cycle_start:] + [key]
            raise DependencyCycleError(cycle)
        visited.append(key)

        chain = list(reversed(self.storage.get_resource_chain(resource.id)))
        effective_mask = Permission.NONE

        # Determine relevant subjects (self + inherited roles)
        relevant_subjects: set[tuple[SubjectType, str]] = set()

        if subject_type == SubjectType.USER:
            user = self.storage.get_user(subject_id)
            relevant_subjects.add((SubjectType.USER, user.id))
            for rid in expand_roles(user.role_ids, self.storage.roles):
                relevant_subjects.add((SubjectType.ROLE, rid))
        else:
            for rid in expand_roles([subject_id], self.storage.roles):
                relevant_subjects.add((SubjectType.ROLE, rid))

        for node in chain:
            node_allow = Permission.NONE
            node_deny = Permission.NONE

            for acl in self.storage.iter_acls_for_resource(node.id):
                if (acl.subject_type, acl.subject_id) not in relevant_subjects:
                    continue

                if not self._check_acl_dependencies(acl, context, visited, cache):
                    continue

                node_allow |= acl.allow_mask
                node_deny |= acl.deny_mask

            if node.inherit_mode == InheritMode.MERGE:
                effective_mask |= node_allow
            elif node.inherit_mode == InheritMode.OVERRIDE:
                effective_mask = node_allow
            # INHERIT does nothing (keeps previous effective_mask)

            if node_deny:
                effective_mask &= ~node_deny

        cache[key] = effective_mask
        visited.pop()
        return effective_mask

    def _check_acl_dependencies(
        self,
        acl: AclEntry,
        context: T | None,
        visited: list[tuple[str, str, str]],
        cache: dict[tuple[str, str, str], Permission],
    ) -> bool:
        if not acl.dependencies:
            return True

        for dep in acl.dependencies:
            dep_res = self.storage.get_resource(dep.resource_id)
            dep_mask = self._calc_permissions_for_subject(
                dep.subject_type, dep.subject_id, dep_res, context, visited, cache
            )
            if (dep_mask & dep.required_mask) != dep.required_mask:
                return False
        return True



from functools import wraps
from typing import Callable, TypeVar, Concatenate, ParamSpec

from arclet.cithun import Permission, User
from arclet.cithun.builtins import System

monitor = System("check_monitor.json")

T = TypeVar("T")
P = ParamSpec("P")

user = monitor.create_user("user:cithun", "cithun")


def require(
    path: str, default_available: bool = False
) -> Callable[[Callable[P, T]], Callable[Concatenate[User, P], T]]:
    def decorator(func: Callable[P, T]) -> Callable[Concatenate[User, P], T]:
        @wraps(func)
        def wrapper(usr: User, *args: P.args, **kwargs: P.kwargs) -> T:
            if monitor.test(user, path, Permission.AVAILABLE):
                return func(*args, **kwargs)
            else:
                raise PermissionError(f"Permission denied for {usr.name} to access {path}")

        return wrapper

    monitor.define(path)
    return decorator


@require("foo.bar.baz")
def alice():
    return "alice"


@require("foo.bar.baz.qux")
def bob():
    return "bob"


@require("foo.bar.qux")
def caven():
    return "caven"


@require("foo.bar.baz.quux")
def dale():
    return "dale"


monitor.suset(user, "foo.bar.*", Permission(7))
assert alice(user) == "alice"
assert bob(user) == "bob"  # target node's parent is available
assert dale(user) == "dale"

try:
    caven(user)
except PermissionError as e:
    # raise PermissionError as caven's target node is not in the available path
    assert str(e) == "Permission denied for cithun to access foo.bar.qux"

monitor.suset(user, "foo.bar.qux", Permission.VISIT | Permission.AVAILABLE | Permission.MODIFY)
assert caven(user) == "caven"  # caven as target node is available

monitor.suset(user, "foo.bar.baz.quux", Permission.VISIT)

try:
    dale(user)
except PermissionError as e:
    # raise PermissionError as dale's target node is dependent on the unavailable node
    assert str(e) == "Permission denied for cithun to access foo.bar.baz.quux"

monitor.depend(user, "foo.bar.qux", user, "foo.bar.baz.qux", Permission.VISIT)
monitor.suset(user, "foo.bar.baz.qux", Permission.VISIT)
try:
    caven(user)
except PermissionError as e:
    # raise PermissionError as caven's target node is dependent on the unavailable node
    assert str(e) == "Permission denied for user:cithun to access foo.bar.qux"

print(monitor.resource_tree())
print(monitor.permission_on(user))



from arclet.cithun import Permission
from arclet.cithun.builtins import DBSystem

system = DBSystem("data.db")

with system.transaction():
    AUTH_1 = system.create_role("ROLE_AUTH_1", "AUTH_1")
    AUTH_2 = system.create_role("ROLE_AUTH_2", "AUTH_2")
    AUTH_3 = system.create_role("ROLE_AUTH_3", "AUTH_3")
    system.inherit(AUTH_2, AUTH_1)
    system.inherit(AUTH_3, AUTH_2)

    alice = system.create_user("alice", "Alice")
    auths = system.create_track("Authority")
    system.extend_track(auths, [AUTH_1, AUTH_2, AUTH_3], ["member", "admin", "owner"])

    system.assign(AUTH_1, "app.data", Permission.AVAILABLE)
    system.assign(AUTH_2, "app.data", Permission.VISIT)
    system.assign(AUTH_3, "app.data", Permission.MODIFY)

    system.set_user_track_level(alice, auths, -1)

    system.promote_track(alice, auths)
    assert system.get_user_track_level(alice, auths).level_name == "member"  # type: ignore
    assert system.test(alice, "app.data", Permission.AVAILABLE)
    assert not system.test(alice, "app.data", Permission.VISIT)
    system.promote_track(alice, auths)
    assert system.get_user_track_level(alice, auths).level_name == "admin"  # type: ignore
    assert system.test(alice, "app.data", Permission.VISIT)
    assert not system.test(alice, "app.data", Permission.MODIFY)
    system.promote_track(alice, auths)
    assert system.get_user_track_level(alice, auths).level_name == "owner"  # type: ignore
    assert system.test(alice, "app.data", Permission.MODIFY)
    #
    # system.assign(AUTH_4, "app.secret", Permission.AVAILABLE | Permission.VISIT)
    # system.assign(alice, "app.data", Permission.MODIFY)
    # system.assign(bob, "app.config", Permission.MODIFY)
    #
    # system.depend(alice, "app.data", AUTH_4, "app.secret", Permission.VISIT)
    # system.depend(bob, "app.config", alice, "app.data", Permission.VISIT)
    #
    #
    system.assign(AUTH_2, "command.foo", Permission.AVAILABLE)
    assert system.test(alice, "command.foo", Permission.AVAILABLE)



from pathlib import Path

from arclet.cithun import Permission, User, ResourceNode
from arclet.cithun.builtins import System

monitor = System("./function_monitor.json")
monitor.define("foo.bar.baz")
monitor.define("foo.bar.baz.qux")
monitor.define("command.test.sub")
monitor.define("command.test1.sub1")


# @monitor.engine.register_strategy
# def auth_handler(
#     user: User,
#     resource: ResourceNode,
#     context: dict | None,
#     current_mask: int,
#     permission_lookup
# ):
#     if not context:
#         return current_mask
#     if resource.id == "command.foo" and context["role"] == "owner":
#         return Permission(7)
#     return current_mask
@monitor.attach("command.foo")
def command_foo_handler(user: User, context: dict | None, current_mask: int, permission_lookup) -> int:
    if context and context.get("role") == "owner":
        return Permission(7)
    return current_mask
# @monitor.attach(lambda pat: pat.startswith("auth."))
# def auth_handler(node: str, owner, state) -> bool:
#     level = int(node.split(".")[-1])
#     if level == 4:
#         return any(ow.name == "owner" for ow in owner.inherits)
#     if level == 3:
#         return any(ow.name == "admin" for ow in owner.inherits)
#     return True


with monitor.transaction():
    default = monitor.create_role("role:default", "Default")
    admin = monitor.create_role("role:admin", "Administrator")
    monitor.inherit(admin, default)

    monitor.suset(admin, "foo.bar.baz", Permission.VISIT | Permission.AVAILABLE | Permission.MODIFY)

    user = monitor.create_user("user:cithun", "cithun")
    monitor.inherit(user, default)
    monitor.inherit(user, admin)

    # monitor.run_attach(user, NodeState("vma"), {})

    assert monitor.test(user, "foo.bar.baz", Permission.VISIT)
    assert not monitor.test(user, "foo.bar.baz.qux", Permission.AVAILABLE)

    monitor.chmod(user, "foo.bar.baz.qux", "a=vm")
    assert monitor.get(user, "foo.bar.baz.qux") == Permission.VISIT | Permission.MODIFY

    monitor.chmod(user, "foo.bar.baz.qux", "a+x")
    assert monitor.get(user, "foo.bar.baz.qux") == Permission.VISIT | Permission.MODIFY | Permission.AVAILABLE

    monitor.chmod(user, "foo.bar.baz.qux", "a-r")
    assert monitor.suget(user, "foo.bar.baz.qux") == Permission.MODIFY | Permission.AVAILABLE

    monitor.chmod(admin, "foo.bar.baz", "a=rx")
    try:
        monitor.set(user, user, "foo.bar.baz.qux", Permission.VISIT | Permission.MODIFY)
    except PermissionError as e:
        print(e)  # Permission denied as "foo.bar.baz" is not modifiable

    monitor.suset(user, "command.test", Permission(5))
    monitor.suset(user, "command.test.*", Permission(5))
    assert monitor.get(user, "command.test.sub") == Permission(5)

    monitor.assign(default, "command.foo", Permission.AVAILABLE)
    assert monitor.test(user, "command.foo", Permission.AVAILABLE)
    assert not monitor.test(user, "command.foo", Permission.VISIT)
    assert monitor.test(user, "command.foo", Permission.MODIFY, context={"role": "owner"})

print(monitor.resource_tree())
print(monitor.permission_on(user))



from __future__ import annotations

import fnmatch
from collections.abc import Callable, Iterable
from itertools import zip_longest
from re import Pattern

from arclet.cithun.config import Config
from arclet.cithun.model import (
    AclDependency,
    AclEntry,
    InheritMode,
    Permission,
    ResourceNode,
    Role,
    Track,
    TrackLevel,
    User,
)


class AsyncStore:

    def __init__(self):
        self.resources: dict[str, ResourceNode] = {}
        self.users: dict[str, User] = {}
        self.roles: dict[str, Role] = {}
        self.acls: list[AclEntry] = []
        self.tracks: dict[str, Track] = {}

    async def _add_resource(self, res: ResourceNode):
        """添加资源节点。

        Args:
            res (ResourceNode): 资源节点对象。
        """
        self.resources[res.id] = res

    async def get_resource(self, rid: str) -> ResourceNode:
        """获取资源节点。

        Args:
            rid (str): 资源 ID。

        Returns:
            ResourceNode: 资源节点对象。
        """
        return self.resources[rid]

    async def get_resource_chain(self, rid: str) -> list[ResourceNode]:
        """获取从当前节点到根节点的资源链。

        Args:
            rid (str): 资源 ID。

        Returns:
            list[ResourceNode]: 资源节点列表，顺序为 [current, parent, ..., root]。
        """
        chain = []
        current = self.resources.get(rid)
        while current:
            chain.append(current)
            if not current.parent_id:
                break
            current = self.resources.get(current.parent_id)
        return chain

    async def define(
        self,
        path: str,
        inherit_mode: InheritMode | None = None,
        type_: str = "GENERIC",
    ) -> ResourceNode:
        """定义资源节点。如果路径上的中间节点不存在，会自动创建为 DIR 类型。

        Args:
            path (str): 资源路径。
            inherit_mode (InheritMode | None, optional): 继承模式。默认为 None。
            type_ (str, optional): 资源类型。默认为 "GENERIC"。

        Returns:
            ResourceNode: 定义的资源节点。

        Raises:
            ValueError: 当 path 为空时抛出。
        """
        parts = [p for p in path.strip(Config.NODE_SEPARATOR).split(Config.NODE_SEPARATOR) if p]
        if not parts:
            raise ValueError("path must not be empty")

        current_parent_id: str | None = None
        node = None

        for i, part in enumerate(parts):
            current_id = f"{current_parent_id}{Config.NODE_SEPARATOR}{part}" if current_parent_id else part
            is_last = i == len(parts) - 1

            node = self.resources.get(current_id)
            if not node:
                node = ResourceNode(
                    id=current_id,
                    name=part,
                    parent_id=current_parent_id,
                    inherit_mode=(inherit_mode or InheritMode.OVERRIDE) if is_last else InheritMode.MERGE,
                    type=type_ if is_last else "DIR",
                )
                await self._add_resource(node)
            elif is_last:
                # 已存在，更新属性
                if inherit_mode is not None:
                    node.inherit_mode = inherit_mode
                node.type = type_
            else:
                # 作为中间节点，确保类型为 DIR，继承模式为 MERGE
                node.inherit_mode = InheritMode.MERGE
                node.type = "DIR"
            current_parent_id = current_id

        return node  # type: ignore

    async def glob_resources(self, pattern: str) -> list[ResourceNode]:
        """使用 glob 模式匹配资源。

        Args:
            pattern (str): glob 模式字符串。

        Returns:
            list[ResourceNode]: 匹配的资源节点列表。
        """
        pattern = pattern.strip(Config.NODE_SEPARATOR)
        result = []
        for res in self.resources.values():
            rid = res.id.strip(Config.NODE_SEPARATOR)
            if fnmatch.fnmatch(rid, pattern):
                result.append(res)
        return result

    async def match_resources(self, pattern: Callable[[str], bool]) -> list[ResourceNode]:
        """使用回调函数匹配资源。

        Args:
            pattern (Callable[[str], bool]): 匹配函数，接受资源 ID 返回布尔值。

        Returns:
            list[ResourceNode]: 匹配的资源节点列表。
        """
        result = []
        for res in self.resources.values():
            rid = res.id.strip(Config.NODE_SEPARATOR)
            if pattern(rid):
                result.append(res)
        return result

    async def assign(
        self,
        subject: User | Role,
        resource_path: str | Callable[[str], bool] | Pattern[str],
        allow_mask: Permission,
        deny_mask: Permission = Permission.NONE,
    ) -> None:
        """为指定 subject 在资源上分配 ACL。

        若 resource_path 中包含 '*'，则按 glob 匹配所有已存在资源，对每个资源应用同一个策略；
        否则对单一资源进行操作。

        注意：
        - assign 只会“创建新 ACL”，不会更新已有 ACL；
        - 对已经存在主 ACL 的 subject+resource，应通过 suset/set 来修改。

        Args:
            subject (User | Role): 目标主体。
            resource_path (str | Callable[[str], bool] | Pattern[str]): 资源路径或匹配模式。
            allow_mask (Permission): 允许权限掩码。
            deny_mask (Permission, optional): 拒绝权限掩码。默认为 0。
        """

        if isinstance(resource_path, str):
            resource_path = resource_path.strip(Config.NODE_SEPARATOR)
            if "*" in resource_path or "?" in resource_path or "[" in resource_path:
                matched = await self.glob_resources(resource_path)
            else:
                res = await self.define(resource_path)
                acl = AclEntry(
                    subject_type=subject.type,
                    subject_id=subject.id,
                    resource_id=res.id,
                    allow_mask=allow_mask,
                    deny_mask=deny_mask,
                )
                await self._add_acl(acl)
                return
        elif isinstance(resource_path, Pattern):
            matched = await self.match_resources(lambda t: bool(resource_path.fullmatch(t)))
        else:
            matched = await self.match_resources(resource_path)

        for res in matched:
            if await self.get_primary_acl(subject, res.id):
                continue
            acl = AclEntry(
                subject_type=subject.type,
                subject_id=subject.id,
                resource_id=res.id,
                allow_mask=allow_mask,
                deny_mask=deny_mask,
            )
            await self._add_acl(acl)

    async def depend(
        self,
        target_subject: User | Role,
        target_resource_id: str,
        dep_subject: User | Role,
        dep_resource_path: str,
        required_mask: Permission,
    ) -> AclEntry:
        """添加 ACL 依赖。

        Args:
            target_subject (User | Role): 目标主体。
            target_resource_id (str): 目标资源 ID。
            dep_subject (User | Role): 依赖主体。
            dep_resource_path (str): 依赖资源路径。
            required_mask (Permission): 依赖所需的权限掩码。

        Returns:
            AclEntry: 更新后的目标 ACL 条目。

        Raises:
            ValueError: 当目标 ACL 不存在时抛出。
        """
        target_acl = await self.get_primary_acl(target_subject, target_resource_id)
        if not target_acl:
            raise ValueError("Target ACL does not exist.")
        dep_res = await self.define(dep_resource_path)
        dep = AclDependency(
            subject_type=dep_subject.type,
            subject_id=dep_subject.id,
            resource_id=dep_res.id,
            required_mask=required_mask,
        )
        target_acl.dependencies.append(dep)
        return target_acl

    def _ensure_user(self, user: User) -> User:
        if user.id in self.users:
            return user
        self.users[user.id] = user
        return user

    def _ensure_role(self, role: Role) -> Role:
        if role.id in self.roles:
            return role
        self.roles[role.id] = role
        return role

    async def inherit(self, child: User | Role, parent: Role):
        """设置继承关系。

        Args:
            child (User | Role): 子主体。
            parent (Role): 父主体。

        Raises:
            ValueError: 当继承关系不合法（如 Role 继承 User）时抛出。
        """
        if isinstance(child, Role):
            child_role = self._ensure_role(child)
            self._ensure_role(parent)
            if parent.id not in child_role.parent_role_ids:
                child_role.parent_role_ids.append(parent.id)
        else:
            user = self._ensure_user(child)
            self._ensure_role(parent)
            if parent.id not in user.role_ids:
                user.role_ids.append(parent.id)

    async def create_user(self, uid: str, name: str) -> User:
        """创建用户。

        Args:
            uid (str): 用户 ID。
            name (str): 用户名称。

        Returns:
            User: 创建的用户对象。
        """
        user = User(id=uid, name=name)
        self.users[user.id] = user
        return user

    async def create_role(self, rid: str, name: str) -> Role:
        """创建角色。

        Args:
            rid (str): 角色 ID。
            name (str): 角色名称。

        Returns:
            Role: 创建的角色对象。
        """
        role = Role(id=rid, name=name)
        self.roles[role.id] = role
        return role

    async def get_user(self, uid: str) -> User:
        """获取用户。

        Args:
            uid (str): 用户 ID。

        Returns:
            User: 用户对象。
        """
        return self.users[uid]

    async def get_role(self, rid: str) -> Role:
        """获取角色。

        Args:
            rid (str): 角色 ID。

        Returns:
            Role: 角色对象。
        """
        return self.roles[rid]

    async def _add_acl(self, acl: AclEntry):
        self.acls.append(acl)

    async def get_acl(self, subject: User | Role, resource_id: str) -> list[AclEntry]:
        """获取指定主体在指定资源上的所有 ACL。

        Args:
            subject (User | Role): 主体。
            resource_id (str): 资源 ID。

        Returns:
            list[AclEntry]: ACL 列表。
        """
        return [
            acl
            for acl in self.acls
            if acl.subject_type == subject.type and acl.subject_id == subject.id and acl.resource_id == resource_id
        ]

    async def get_primary_acl(
        self,
        subject: User | Role,
        resource_id: str,
    ) -> AclEntry | None:
        """获取指定主体在指定资源上的主 ACL（第一个匹配的 ACL）。

        Args:
            subject (User | Role): 主体。
            resource_id (str): 资源 ID。

        Returns:
            AclEntry | None: 主 ACL 条目，若不存在则返回 None。
        """
        return next(
            (
                acl
                for acl in self.acls
                if acl.subject_type == subject.type and acl.subject_id == subject.id and acl.resource_id == resource_id
            ),
            None,
        )

    async def iter_acls_for_resource(self, resource_id: str) -> Iterable[AclEntry]:
        """迭代指定资源的所有 ACL。

        Args:
            resource_id (str): 资源 ID。

        Returns:
            Iterable[AclEntry]: ACL 迭代器。
        """
        return (acl for acl in self.acls if acl.resource_id == resource_id)

    async def create_track(self, tid: str, name: str | None = None) -> Track:
        """创建或获取 Track。

        Args:
            tid (str): Track ID。
            name (str | None, optional): Track 名称。默认为 None。

        Returns:
            Track: Track 对象。
        """
        if tid in self.tracks:
            return self.tracks[tid]
        track = Track(id=tid, name=name or tid)
        self.tracks[track.id] = track
        return track

    async def _add_track(self, track: Track):
        """添加 Track。

        Args:
            track (Track): Track 对象。
        """
        self.tracks[track.id] = track

    def get_track(self, tid: str) -> Track:
        """获取 Track。

        Args:
            tid (str): Track ID。

        Returns:
            Track: Track 对象。
        """
        return self.tracks[tid]

    async def add_track_level(self, track: Track, role: Role, name: str | None = None) -> None:
        """向 Track 添加一个等级。

        Args:
            track (Track): Track 对象。
            role (Role): 对应的角色。
            name (str | None, optional): 等级名称。默认为 None。
        """
        track.levels.append(
            TrackLevel(
                role_id=role.id,
                level_name=name or role.name,
            )
        )

    async def extend_track(self, track: Track, roles: Iterable[Role], names: Iterable[str] | None = None) -> None:
        """批量向 Track 添加等级。

        Args:
            track (Track): Track 对象。
            roles (Iterable[Role]): 角色列表。
            names (Iterable[str] | None, optional): 等级名称列表。默认为 None。
        """
        for role, name in zip_longest(roles, names or []):
            await self.add_track_level(track, role, name)

    async def insert_track_level(self, track: Track, index: int, role: Role, name: str | None = None) -> None:
        """在指定位置插入 Track 等级。

        Args:
            track (Track): Track 对象。
            index (int): 插入位置索引。
            role (Role): 对应的角色。
            name (str | None, optional): 等级名称。默认为 None。
        """
        track.levels.insert(
            index,
            TrackLevel(
                role_id=role.id,
                level_name=name or role.name,
            ),
        )

    def get_user_track_level(self, user: User, track: Track) -> TrackLevel | None:
        """获取用户在某个 Track 上的当前等级。

        Args:
            user (User): 用户对象。
            track (Track): Track 对象。
        Returns:
            TrackLevel | None: 当前等级对象，若用户不在该 Track 上则返回 None。
        """
        for level in track.levels:
            if level.role_id in user.role_ids:
                return level
        return None

    async def set_user_track_level(self, user: User, track: Track, level_index: int) -> None:
        """将用户在某个 Track 上设置到指定等级。

        会清理该用户在该 Track 上已有的其他角色，并赋予新等级对应的角色。

        Args:
            user (User): 用户对象。
            track (Track): Track 对象。
            level_index (int): 目标等级索引。

        Raises:
            ValueError: 当 Track 没有等级或索引无效时抛出。
        """
        levels = track.levels
        if not levels:
            raise ValueError("Track has no levels.")
        if level_index >= len(levels):
            raise ValueError("Invalid level index.")
        track_role_ids = {level.role_id for level in levels}
        user.role_ids = [rid for rid in user.role_ids if rid not in track_role_ids]
        if level_index < 0:
            return
        if levels[level_index].role_id not in user.role_ids:
            user.role_ids.append(levels[level_index].role_id)

    async def promote_track(self, user: User, track: Track, step: int = 1):
        """用户在某个 Track 上升级。

        - 当前没有该 Track 角色 -> 为用户挂上最低 level 角色
        - 升级超过最高 level -> 停在最高 level

        Args:
            user (User): 用户对象。
            track (Track): Track 对象。
            step (int, optional): 升级步数。默认为 1。

        Returns:
            int | None: 最终的 level_index。如果 Track 没定义等级则返回 None。
        """
        levels = track.levels
        if not levels:
            return
        current_level_index = -1
        for i, level in enumerate(levels):
            if level.role_id in user.role_ids:
                current_level_index = i
                break

        if current_level_index == -1:
            current_level_index = 0
            await self.set_user_track_level(user, track, current_level_index)
            return current_level_index

        new_level_index = min(current_level_index + step, len(levels) - 1)
        await self.set_user_track_level(user, track, new_level_index)
        return new_level_index

    async def demote_track(self, user: User, track: Track, step: int = 1):
        """用户在某个 Track 上降级。

        - 当前没有该 Track 角色 -> 无视
        - 降级超过最低 level -> 停在最低 level

        Args:
            user (User): 用户对象。
            track (Track): Track 对象。
            step (int, optional): 降级步数。默认为 1。

        Returns:
            int | None: 最终的 level_index。如果 Track 没定义等级或用户不在 Track 中则返回 None。
        """
        levels = track.levels
        if not levels:
            return
        current_level_index = -1
        for i, level in enumerate(levels):
            if level.role_id in user.role_ids:
                current_level_index = i
                break

        if current_level_index == -1:
            return

        new_level_index = max(current_level_index - step, 0)
        await self.set_user_track_level(user, track, new_level_index)
        return new_level_index

    def resource_tree(self) -> str:
        """生成资源树的字符串表示。

        Returns:
            str: 资源树字符串。
        """
        lines = ["/"]

        def _format_node(node: ResourceNode, prefix: str, is_last: bool):
            children = [n for n in self.resources.values() if n.parent_id == node.id]
            lines.append(f"{prefix}{'└─ ' if is_last else '├─ '}{node.name}{'/' if children else ''}")
            for i, child in enumerate(children):
                _format_node(child, prefix + ("   " if is_last else "│  "), i == len(children) - 1)

        roots = [n for n in self.resources.values() if n.parent_id is None]
        for i, root in enumerate(roots):
            _format_node(root, "", i == len(roots) - 1)
        return "\n".join(lines)

    async def permission_on(self, subject: User | Role):
        """生成指定主体在所有资源上的权限视图。

        Args:
            subject (User | Role): 目标主体。

        Returns:
            str: 权限视图字符串。
        """
        lines = ["/"]

        async def _format_node(node: ResourceNode, prefix: str, is_last: bool):
            acl = await self.get_primary_acl(subject, node.id)
            if acl:
                perm_str = f"(allow: {acl.allow_mask:#}, deny: {acl.deny_mask:#})"
            else:
                perm_str = ""
            children = [n for n in self.resources.values() if n.parent_id == node.id]
            lines.append(f"{prefix}{'└─ ' if is_last else '├─ '}{node.name}{'/' if children else ''} {perm_str}")
            for i, child in enumerate(children):
                await _format_node(child, prefix + ("   " if is_last else "│  "), i == len(children) - 1)

        roots = [n for n in self.resources.values() if n.parent_id is None]
        for i, root in enumerate(roots):
            await _format_node(root, "", i == len(roots) - 1)
        return "\n".join(lines)

    async def update_acl(self, acl: AclEntry, allow_mask: Permission, deny_mask: Permission | None = None) -> None:
        """更新 ACL 条目。

        Args:
            acl (AclEntry): 要更新的 ACL 条目。
            allow_mask (int): 新的允许权限掩码。
            deny_mask (int | None, optional): 新的拒绝权限掩码。
        """
        acl.allow_mask = allow_mask
        if deny_mask is not None:
            acl.deny_mask = deny_mask



from arclet.cithun.model import Permission


assert Permission.parse("a=7") == (7, "=", False)
assert Permission.parse("d+vma") == (Permission.VISIT | Permission.MODIFY | Permission.AVAILABLE, "+", True)
assert Permission.parse("av") == (Permission.VISIT, "=", False)
assert Permission.parse("a-m") == (Permission.MODIFY, "-", False)
assert Permission.parse("5") == (5, "=", False)
assert Permission.parse("a") == (Permission.AVAILABLE, "=", False)



from contextlib import contextmanager

from arclet.cithun import PermissionEngine, PermissionExecutor, PermissionService

from .attach import Attacher
from .database import SimpleDatabaseStore
from .json import JsonStore


class System(Attacher, JsonStore, PermissionService[dict], PermissionExecutor[dict]):
    def __init__(self, path):
        JsonStore.__init__(self, path)
        PermissionService.__init__(self, engine=PermissionEngine(), storage=self)
        PermissionExecutor.__init__(self, self, self)
        Attacher.__init__(self, self.engine)

    @contextmanager
    def transaction(self):
        yield
        self.save()


class DBSystem(Attacher, SimpleDatabaseStore, PermissionService[dict], PermissionExecutor[dict]):
    def __init__(self, path):
        SimpleDatabaseStore.__init__(self, path)
        PermissionService.__init__(self, engine=PermissionEngine(), storage=self)
        PermissionExecutor.__init__(self, self, self)
        Attacher.__init__(self, self.engine)

    @contextmanager
    def transaction(self):
        self.load()
        yield
        self.save()



from __future__ import annotations

import os
import sqlite3
from pathlib import Path

from arclet.cithun.model import AclDependency, AclEntry, Permission, ResourceNode, Role, Track, TrackLevel, User
from arclet.cithun.store import BaseStore


class SimpleDatabaseStore(BaseStore):
    def __init__(self, file: os.PathLike[str]):
        file = Path(file)
        self.file = file
        self.conn = sqlite3.connect(self.file)
        super().__init__()
        self.ensure_table()

    def ensure_table(self):
        sql = """\
        CREATE TABLE IF NOT EXISTS users (
            id TEXT PRIMARY KEY,
            name TEXT NOT NULL
        );
        CREATE TABLE IF NOT EXISTS roles (
            id TEXT PRIMARY KEY,
            name TEXT NOT NULL
        );

        CREATE TABLE IF NOT EXISTS user_roles (
            user_id TEXT,
            role_id TEXT,
            PRIMARY KEY (user_id, role_id),
            FOREIGN KEY (user_id) REFERENCES users(id) ON DELETE CASCADE,
            FOREIGN KEY (role_id) REFERENCES roles(id) ON DELETE CASCADE
        );

        CREATE TABLE IF NOT EXISTS role_inherits (
            role_id TEXT,
            parent_role_id TEXT,
            PRIMARY KEY (role_id, parent_role_id),
            FOREIGN KEY (role_id) REFERENCES roles(id) ON DELETE CASCADE,
            FOREIGN KEY (parent_role_id) REFERENCES roles(id) ON DELETE CASCADE
        );

        CREATE TABLE IF NOT EXISTS resources (
            id TEXT PRIMARY KEY,
            name TEXT NOT NULL,
            parent_id TEXT NULL,
            inherit_mode INTEGER DEFAULT 0,  -- 0: INHERIT, 1: MERGE, 2: OVERRIDE
            type TEXT NOT NULL DEFAULT 'GENERIC',  -- FILE / DIR / PROJECT / etc.
            FOREIGN KEY (parent_id) REFERENCES resources(id) ON DELETE CASCADE
        );
        CREATE INDEX IF NOT EXISTS idx_resources_parent_id ON resources(parent_id);

        CREATE TABLE IF NOT EXISTS acls (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            subject_type TEXT NOT NULL,  -- 'USER' or 'ROLE'
            subject_id TEXT NOT NULL,
            resource_id TEXT NOT NULL,
            allow_mask INTEGER NOT NULL,
            deny_mask INTEGER DEFAULT 0,
            FOREIGN KEY (resource_id) REFERENCES resources(id) ON DELETE CASCADE
        );

        CREATE INDEX IF NOT EXISTS idx_acls_resource_id ON acls(resource_id);
        CREATE INDEX IF NOT EXISTS idx_acls_subject ON acls(subject_type, subject_id);

        CREATE TABLE IF NOT EXISTS acl_dependencies (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            acl_id INTEGER NOT NULL,
            dep_subject_type TEXT NOT NULL,
            dep_subject_id TEXT NOT NULL,
            dep_resource_id TEXT NOT NULL,
            required_mask INTEGER NOT NULL,
            FOREIGN KEY (acl_id) REFERENCES acls(id) ON DELETE CASCADE,
            FOREIGN KEY (dep_resource_id) REFERENCES resources(id) ON DELETE CASCADE
        );

        CREATE INDEX IF NOT EXISTS idx_acl_deps_acl_id  ON acl_dependencies(acl_id);
        CREATE INDEX IF NOT EXISTS idx_acl_deps_dep_subject ON acl_dependencies(dep_subject_type, dep_subject_id);
        CREATE INDEX IF NOT EXISTS idx_acl_deps_dep_resource ON acl_dependencies(dep_resource_id);

        CREATE TABLE IF NOT EXISTS tracks (
            id TEXT PRIMARY KEY,
            name TEXT NOT NULL
        );
        CREATE TABLE IF NOT EXISTS track_levels (
            `index` INTEGER,
            track_id TEXT,
            role_id TEXT,
            level_name TEXT NOT NULL,
            PRIMARY KEY (track_id, role_id),
            FOREIGN KEY (track_id) REFERENCES tracks(id) ON DELETE CASCADE,
            FOREIGN KEY (role_id) REFERENCES roles(id) ON DELETE CASCADE
        );
        CREATE INDEX IF NOT EXISTS idx_track_levels_track_id ON track_levels(track_id);
        CREATE INDEX IF NOT EXISTS idx_track_levels_role_id ON track_levels(role_id);
        """
        self.conn.executescript(sql)
        self.conn.commit()

    def create_user(self, uid: str, name: str) -> User:
        if uid in self.users:
            return self.users[uid]
        user = User(uid, name)
        self.users[uid] = user
        cursor = self.conn.cursor()
        cursor.execute("INSERT INTO users (id, name) VALUES (?, ?);", (user.id, user.name))
        self.conn.commit()
        return user

    def create_role(self, rid: str, name: str) -> Role:
        if rid in self.roles:
            return self.roles[rid]
        role = Role(rid, name)
        self.roles[rid] = role
        cursor = self.conn.cursor()
        cursor.execute("INSERT INTO roles (id, name) VALUES (?, ?);", (role.id, role.name))
        self.conn.commit()
        return role

    def create_track(self, tid: str, name: str | None = None) -> Track:
        if tid in self.tracks:
            return self.tracks[tid]
        track = Track(tid, name or tid)
        self.tracks[tid] = track
        cursor = self.conn.cursor()
        cursor.execute("INSERT INTO tracks (id, name) VALUES (?, ?);", (track.id, track.name))
        self.conn.commit()
        return track

    def _add_acl(self, acl: AclEntry):
        target_acl = next(
            (
                i
                for i in self.acls
                if i.subject_type == acl.subject_type
                and i.subject_id == acl.subject_id
                and i.resource_id == acl.resource_id
            ),
            None,
        )
        if target_acl:
            return target_acl
        cursor = self.conn.cursor()
        cursor.execute(
            "INSERT INTO acls (subject_type, subject_id, resource_id, allow_mask, deny_mask) "
            "VALUES (?, ?, ?, ?, ?);",
            (acl.subject_type.value, acl.subject_id, acl.resource_id, acl.allow_mask, acl.deny_mask),
        )
        self.conn.commit()
        self.acls.append(acl)

    def _add_resource(self, res: ResourceNode):
        if res.id in self.resources:
            return self.resources[res.id]
        cursor = self.conn.cursor()
        cursor.execute(
            "INSERT INTO resources (id, name, parent_id, inherit_mode, type) " "VALUES (?, ?, ?, ?, ?);",
            (res.id, res.name, res.parent_id, res.inherit_mode.value, res.type),
        )
        self.conn.commit()
        self.resources[res.id] = res
        return res

    def depend(
        self,
        target_subject: User | Role,
        target_resource_id: str,
        dep_subject: User | Role,
        dep_resource_path: str,
        required_mask: Permission,
    ) -> AclEntry:
        target_acl = self.get_primary_acl(target_subject, target_resource_id)
        if not target_acl:
            raise ValueError("Target ACL does not exist.")
        dep_res = self.define(dep_resource_path)
        dep = AclDependency(
            subject_type=dep_subject.type,
            subject_id=dep_subject.id,
            resource_id=dep_res.id,
            required_mask=required_mask,
        )
        if dep in target_acl.dependencies:
            return target_acl
        target_acl.dependencies.append(dep)

        cursor = self.conn.cursor()
        cursor.execute(
            "SELECT id FROM acls WHERE subject_type = ? AND subject_id = ? AND resource_id = ?;",
            (target_acl.subject_type.value, target_acl.subject_id, target_acl.resource_id),
        )
        row = cursor.fetchone()
        if row is None:
            raise ValueError("ACL entry not found after creation.")
        acl_id = row[0]

        cursor.execute(
            "INSERT INTO acl_dependencies (acl_id, dep_subject_type, dep_subject_id, "
            "dep_resource_id, required_mask) VALUES (?, ?, ?, ?, ?);",
            (acl_id, dep.subject_type.value, dep.subject_id, dep.resource_id, dep.required_mask),
        )
        self.conn.commit()
        return target_acl

    def inherit(self, child: User | Role, parent: Role):
        if isinstance(child, Role):
            child_role = self._ensure_role(child)
            self._ensure_role(parent)
            if parent.id not in child_role.parent_role_ids:
                cursor = self.conn.cursor()
                cursor.execute(
                    "INSERT INTO role_inherits (role_id, parent_role_id) VALUES (?, ?);", (child_role.id, parent.id)
                )
                self.conn.commit()
                child_role.parent_role_ids.append(parent.id)
        else:
            user = self._ensure_user(child)
            self._ensure_role(parent)
            if parent.id not in user.role_ids:
                cursor = self.conn.cursor()
                cursor.execute("INSERT INTO user_roles (user_id, role_id) VALUES (?, ?);", (user.id, parent.id))
                self.conn.commit()
                user.role_ids.append(parent.id)

    def add_track_level(self, track: Track, role: Role, name: str | None = None) -> None:
        level = TrackLevel(role.id, name or role.name)
        if level in track.levels:
            return
        track.levels.append(level)
        cursor = self.conn.cursor()
        cursor.execute(
            "INSERT INTO track_levels (`index`, track_id, role_id, level_name) VALUES (?, ?, ?, ?);",
            (len(track.levels) - 1, track.id, role.id, level.level_name),
        )
        self.conn.commit()

    def insert_track_level(self, track: Track, index: int, role: Role, name: str | None = None) -> None:
        level = TrackLevel(role.id, name or role.name)
        if level in track.levels:
            return
        track.levels.insert(index, level)
        cursor = self.conn.cursor()
        cursor.execute("DELETE FROM track_levels WHERE track_id = ?;", (track.id,))
        for idx, lvl in enumerate(track.levels):
            cursor.execute(
                "INSERT INTO track_levels (`index`, track_id, role_id, level_name) VALUES (?, ?, ?, ?);",
                (idx, track.id, lvl.role_id, lvl.level_name),
            )
        self.conn.commit()

    def update_acl(self, acl: AclEntry, allow_mask: Permission, deny_mask: Permission | None = None) -> None:
        acl.allow_mask = allow_mask
        if deny_mask is not None:
            acl.deny_mask = deny_mask
        cursor = self.conn.cursor()
        cursor.execute(
            "UPDATE acls SET allow_mask = ?, deny_mask = ? "
            "WHERE subject_type = ? AND subject_id = ? AND resource_id = ?;",
            (acl.allow_mask, acl.deny_mask, acl.subject_type.value, acl.subject_id, acl.resource_id),
        )
        self.conn.commit()

    def load(self):
        cursor = self.conn.cursor()
        cursor.execute("SELECT id, name FROM users;")
        for row in cursor.fetchall():
            user = User(*row, role_ids=[])
            self.users[user.id] = user
        cursor.execute("SELECT id, name FROM roles;")
        for row in cursor.fetchall():
            role = Role(*row, parent_role_ids=[])
            self.roles[role.id] = role
        for user in self.users.values():
            cursor.execute("SELECT role_id FROM user_roles WHERE user_id = ?;", (user.id,))
            role_ids = [r[0] for r in cursor.fetchall()]
            user.role_ids.extend(role_ids)
        for role in self.roles.values():
            cursor.execute("SELECT parent_role_id FROM role_inherits WHERE role_id = ?;", (role.id,))
            parent_role_ids = [r[0] for r in cursor.fetchall()]
            role.parent_role_ids.extend(parent_role_ids)
        cursor.execute("SELECT id, name, parent_id, inherit_mode, type FROM resources;")
        for row in cursor.fetchall():
            resource = ResourceNode(*row)
            self.resources[resource.id] = resource
        cursor.execute("SELECT id, subject_type, subject_id, resource_id, allow_mask, deny_mask FROM acls;")
        acl_ids = []
        for row in cursor.fetchall():
            acl = AclEntry(*row[1:], dependencies=[])
            self.acls.append(acl)
            acl_ids.append(row[0])
        for acl, acl_id in zip(self.acls, acl_ids):
            cursor.execute(
                "SELECT dep_subject_type, dep_subject_id, dep_resource_id, required_mask "
                "FROM acl_dependencies WHERE acl_id = ?;",
                (acl_id,),
            )
            for dep_row in cursor.fetchall():
                dep = AclDependency(*dep_row)
                acl.dependencies.append(dep)
        cursor.execute("SELECT id, name FROM tracks;")
        for row in cursor.fetchall():
            track = Track(*row, levels=[])
            self.tracks[track.id] = track
        for track in self.tracks.values():
            cursor.execute(
                "SELECT role_id, level_name FROM track_levels WHERE track_id = ? ORDER BY `index` ASC;", (track.id,)
            )
            for level_row in cursor.fetchall():
                level = TrackLevel(*level_row)
                track.levels.append(level)

    def save(self):
        pass



from __future__ import annotations

import re
from dataclasses import dataclass, field
from enum import Enum, IntFlag, auto
from typing import ClassVar


class Permission(IntFlag):
    """权限标志位枚举。"""

    NONE = 0
    AVAILABLE = auto()
    MODIFY = auto()
    VISIT = auto()

    @classmethod
    def _missing_(cls, value):
        if isinstance(value, str):
            value = value.upper()
            mask = cls.NONE
            for char in value:
                if char == "A" or char == "X":
                    mask |= cls.AVAILABLE
                elif char == "M" or char == "W":
                    mask |= cls.MODIFY
                elif char == "V" or char == "R":
                    mask |= cls.VISIT
                elif char == "*":
                    mask |= cls.AVAILABLE | cls.MODIFY | cls.VISIT
                elif char == "-":
                    continue
                else:
                    raise ValueError(f"Invalid permission character: {char}")
            return cls(mask)
        return super()._missing_(value)

    def __format__(self, format_spec):
        if format_spec == "#":
            flags = []
            for flag, char in (
                (Permission.VISIT, "v"),
                (Permission.MODIFY, "m"),
                (Permission.AVAILABLE, "a"),
            ):
                if self & flag:
                    flags.append(char)
                else:
                    flags.append("-")
            return "".join(flags)
        return super().__format__(format_spec)

    @classmethod
    def parse(cls, expr: str):
        """解析权限表达式为掩码。
        Args:
            expr (str): 权限表达式。形如 [target][op][flags].
        Returns:
            tuple[Permission, str, bool]: (mask, mode, deny)
        """
        expr = expr.strip().lower()
        if not expr:
            raise ValueError("Empty permission expression")
        if len(expr) == 1:
            return cls(int(expr) if expr.isdigit() else expr), "=", False
        mat = re.fullmatch(r"(?P<target>[ad])?(?P<op>[=+-])?(?P<flags>(?:[*0-7]|[vmarwx]+))", expr)
        if not mat:
            raise ValueError(f"Invalid permission expression: {expr!r}")
        deny = (mat.groupdict()["target"] or "a") == "d"
        mode = mat.groupdict()["op"] or "="
        flags_str = mat.groupdict()["flags"]
        return cls(int(flags_str) if flags_str.isdigit() else flags_str), mode, deny


class InheritMode(str, Enum):
    """权限继承模式。"""

    INHERIT = "INHERIT"
    """完全继承父节点"""

    MERGE = "MERGE"
    """父 + 子合并"""

    OVERRIDE = "OVERRIDE"
    """只看当前节点"""


@dataclass
class ResourceNode:
    """资源节点。"""

    id: str
    name: str
    parent_id: str | None = None
    inherit_mode: InheritMode = InheritMode.MERGE
    type: str = "GENERIC"  # FILE / DIR / PROJECT / etc.


class SubjectType(str, Enum):
    """主体类型（用户或角色）。"""

    USER = "USER"
    ROLE = "ROLE"


@dataclass
class Role:
    """角色。"""

    id: str
    name: str
    parent_role_ids: list[str] = field(default_factory=list)

    type: ClassVar[SubjectType] = SubjectType.ROLE


@dataclass
class User:
    """用户。"""

    id: str
    name: str
    role_ids: list[str] = field(default_factory=list)

    type: ClassVar[SubjectType] = SubjectType.USER


@dataclass(eq=True)
class AclDependency:
    """描述一个 ACL 对“另一个 subject 在某资源上的权限”的依赖。"""

    subject_type: SubjectType
    subject_id: str
    resource_id: str
    required_mask: Permission


@dataclass
class AclEntry:
    """访问控制列表条目。"""

    subject_type: SubjectType
    subject_id: str
    resource_id: str
    allow_mask: Permission
    deny_mask: Permission = Permission.NONE
    dependencies: list[AclDependency] = field(default_factory=list)

    def __repr__(self) -> str:
        return (
            f"AclEntry(subject={self.subject_type.value}:{self.subject_id}, "
            f"resource={self.resource_id}, allow={self.allow_mask!r}, deny={self.deny_mask!r}, "
            f"deps={self.dependencies})"
        )


@dataclass(eq=True)
class TrackLevel:
    """
    Track 中的一个“等级节点”：
    - 对应一个 Role
    - 有一个顺序 level_index（0,1,2,...，index 越大代表等级越高）
    """

    role_id: str
    level_name: str  # 例如：MEMBER, ADMIN, OWNER


@dataclass
class Track:
    """
    一条“轨道”，管理一组有序角色：
    比如：AUTH_1 ~ AUTH_5。
    """

    id: str
    name: str
    levels: list[TrackLevel] = field(default_factory=list)


if __name__ == "__main__":
    print(repr(Permission(7)))
    print(repr(Permission("v-a")))



from __future__ import annotations

from collections.abc import Awaitable, Callable
from typing import Generic, Protocol, TypeVar

from arclet.cithun.model import Permission, ResourceNode, Role, User

T = TypeVar("T")


class AsyncPermissionStrategy(Protocol[T]):
    """权限策略协议。"""

    async def __call__(
        self,
        user: User,
        resource: ResourceNode,
        context: T | None,
        current_mask: Permission,
        permission_lookup: Callable[[User | Role, T | None], Awaitable[Permission]],
    ) -> Permission:
        """执行策略。

        Args:
            user (User): 用户对象。
            resource (ResourceNode): 资源节点。
            context (T, optional): 上下文信息。
            current_mask (Permission): 当前权限掩码。
            permission_lookup (Callable[[User | Role, T | None], Awaitable[Permission]]): 权限查找回调函数。

        Returns:
            Permission: 更新后的权限掩码。
        """
        ...


class AsyncPermissionEngine(Generic[T]):
    """权限引擎，管理和应用权限策略。"""

    def __init__(self):
        self._strategies: list[AsyncPermissionStrategy[T]] = []

    def register_strategy(self, strategy: AsyncPermissionStrategy[T]):
        """注册策略。

        Args:
            strategy (AsyncPermissionStrategy[T]): 策略对象。
        """
        self._strategies.append(strategy)

    async def apply_strategies(
        self,
        user: User,
        resource: ResourceNode,
        context: T | None,
        mask: Permission,
        permission_lookup: Callable[[User | Role, T | None], Awaitable[Permission]],
    ) -> Permission:
        """应用所有注册的策略。

        Args:
            user (User): 用户对象。
            resource (ResourceNode): 资源节点。
            context (T, optional): 上下文信息。
            mask (Permission): 初始权限掩码。
            permission_lookup (Callable[[User | Role, T | None], Awaitable[Permission]]): 权限查找回调函数。

        Returns:
            Permission: 最终权限掩码。
        """
        for s in self._strategies:
            mask = await s(user, resource, context, mask, permission_lookup)
        return mask



from collections.abc import Callable
from functools import reduce
from operator import or_
from typing import TypeAlias, TypeVar, overload

from arclet.cithun import Permission, PermissionEngine, ResourceNode, Role, User

Attach: TypeAlias = "Callable[[User, dict | None, int, Callable[[User | Role, dict | None], Permission]], Permission]"
TAttach = TypeVar("TAttach", bound=Attach)
Attach1: TypeAlias = (
    "Callable[[User, str, dict | None, int, Callable[[User | Role, dict | None], Permission]], Permission]"
)
TAttach1 = TypeVar("TAttach1", bound=Attach1)


class Attacher:
    def __init__(self, engine: PermissionEngine[dict]):
        self.attachs = []
        engine.register_strategy(self._run_attachs)

    def _run_attachs(
        self,
        user: User,
        resource: ResourceNode,
        context: dict | None,
        current_mask: Permission,
        permission_lookup: Callable[[User | Role, dict | None], Permission],
    ) -> Permission:
        results = []
        for pattern, func in self.attachs:
            if pattern(resource.id):
                results.append(func(user, resource.id, context, current_mask, permission_lookup))
        if not results:
            return current_mask
        return reduce(or_, results, current_mask)

    @overload
    def attach(self, pattern: str) -> Callable[[TAttach], TAttach]: ...

    @overload
    def attach(self, pattern: Callable[[str], bool]) -> Callable[[TAttach1], TAttach1]: ...

    def attach(self, pattern):  # type: ignore
        if isinstance(pattern, str):

            def decorator(func: Attach, /):
                self.attachs.append((lambda p: p == pattern, lambda u, _, *args: func(u, *args)))
                return func

            return decorator

        def wrapper(func: Attach1, /):
            self.attachs.append((pattern, func))
            return func

        return wrapper



from arclet.cithun.model import InheritMode as InheritMode
from arclet.cithun.model import Permission as Permission
from arclet.cithun.model import ResourceNode as ResourceNode
from arclet.cithun.model import Role as Role
from arclet.cithun.model import User as User

from .function import AsyncPermissionExecutor as AsyncPermissionExecutor
from .service import AsyncPermissionService as AsyncPermissionService
from .store import AsyncStore as AsyncStore
from .strategy import AsyncPermissionEngine as AsyncPermissionEngine

PE = AsyncPermissionExecutor



class ResourceNotFoundError(FileNotFoundError):
    """资源不存在或父资源不存在时抛出。"""

    pass


class PermissionDeniedError(PermissionError):
    """权限不足时抛出。"""

    pass


class DependencyCycleError(RuntimeError):
    """检测到依赖循环时抛出。"""

    def __init__(self, cycle_nodes: list[tuple[str, str, str]]):
        self.cycle_nodes = cycle_nodes
        msg = "Dependency cycle detected: " + " -> ".join(f"{t}:{sid}@{rid}" for (t, sid, rid) in cycle_nodes)
        super().__init__(msg)
